#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import cv2 as cv

from ..coreten import Tensor, to_tensor
from ._constants import HLS2BGR, HLS2RGB
from ._bgr import bgr2gray, bgr2lab, bgr2hsv, bgr2yuv, bgr2luv


__all__ = [
    'hls2rgb',
    'hls2bgr',
    'hls2lab',
    'hls2gray',
    'hls2hsv',
    'hls2yuv',
    'hls2luv',
]


def _is_hls_image(tens):
    # tens = to_tensor(tens)
    # return tens.is_hls()
    return len(tens.shape) == 3 and tens.shape[-1] == 3


def hls2rgb(tens) -> Tensor:
    r'''
        Converts a HLS Tensor to its RGB version.

    Args:
        tens (Tensor): Valid HLS Tensor

    Returns:
        RGB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_hls_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a HLS Tensor to its RGB counterpart'
        )

    im = cv.cvtColor(tens, HLS2RGB)
    return to_tensor(im, cspace='rgb')


def hls2bgr(tens) -> Tensor:
    r'''
        Converts a HLS Tensor to its BGR version.

    Args:
        tens (Tensor): Valid HLS Tensor

    Returns:
        BGR Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_hls_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a HLS Tensor to its BGR counterpart'
        )

    im = cv.cvtColor(tens, HLS2BGR)
    return to_tensor(im, cspace='bgr')


def hls2gray(tens) -> Tensor:
    r'''
        Converts a HLS Tensor to its Grayscale version.

    Args:
        tens (Tensor): Valid HLS Tensor

    Returns:
        Grayscale Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_hls_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a HLS Tensor to its Grayscale counterpart'
        )

    bgr = hls2bgr(tens)

    im = bgr2gray(bgr)
    return to_tensor(im, cspace='gray')


def hls2hsv(tens) -> Tensor:
    r'''
        Converts a HLS Tensor to its HSV version.

    Args:
        tens (Tensor): Valid HLS Tensor

    Returns:
        HSV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_hls_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a HLS Tensor to its LAB counterpart'
        )

    bgr = hls2bgr(tens)

    im = bgr2hsv(bgr)
    return to_tensor(im, cspace='hsv')


def hls2lab(tens) -> Tensor:
    r'''
        Converts a HLS Tensor to its LAB version.

    Args:
        tens (Tensor): Valid HLS Tensor

    Returns:
        LAB Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_hls_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a HLS Tensor to its LAB counterpart'
        )

    bgr = hls2bgr(tens)

    im = bgr2lab(bgr)
    return to_tensor(im, cspace='lab')


def hls2yuv(tens) -> Tensor:
    r'''
        Converts a HLS Tensor to its YUV version.

    Args:
        tens (Tensor): Valid HLS Tensor

    Returns:
        YUV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_hls_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a HLS Tensor to its YUV counterpart'
        )

    bgr = hls2bgr(tens)

    im = bgr2yuv(bgr)
    return to_tensor(im, cspace='yuv')


def hls2luv(tens) -> Tensor:
    r'''
        Converts a HLS Tensor to its LUV version.

    Args:
        tens (Tensor): Valid HLS Tensor

    Returns:
        YUV Tensor of shape ``(height, width, channels)``

    Raises:
        ValueError: If `tens` is not of shape 3

    '''
    if not _is_hls_image(tens):
        raise ValueError(
            f'Tensor of shape 3 expected. Found shape {len(tens.shape)}. This function converts a HLS Tensor to its LUV counterpart'
        )

    bgr = hls2bgr(tens)

    im = bgr2luv(bgr)
    return to_tensor(im, cspace='luv')



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


#pylint:disable=redefined-outer-name

import os
from time import time
from ..annotations import List, Union

_acceptable_video_formats = ('.mp4', '.avi', '.mov', '.mkv', '.webm')
_acceptable_image_formats = ('.jpg', '.jpeg', '.png', '.bmp', '.tif', '.tiff')

__all__ = [
    'list_images',
    'list_videos',
    'listdir',
    'is_image',
    'is_video',
    'isfile',
    'mkdir',
    'cwd',
    'exists',
    'join',
    'get_size',
    'chdir',
    'osname',
    'abspath',
    'dirname'
]

def list_images(DIR, recursive=True, use_fullpath=False, verbose=True, every : int = 1000) -> List[str]:
    r"""
        Lists all image files within a specific directory (and sub-directories if `recursive=True`)
    
    Args:
        DIR (str): Directory to search for image files
        recursive (bool): Indicate whether to search all subdirectories as well (default = False)
        use_fullpath (bool): Include full filepaths in the returned list (default = False)
    
    Returns:
        image_files (list): List of names (or full filepaths if `use_fullpath=True`) of the image files

    """
    return listdir(DIR=DIR, recursive=recursive, use_fullpath=use_fullpath, ext = _acceptable_image_formats, verbose=verbose, every=every)


def list_videos(DIR, recursive=True, use_fullpath=False, verbose=True, every : int = 1000) -> List[str]:
    r"""
        Lists all video files within a specific directory (and sub-directories if `recursive=True`)
    
    Args:
        DIR (str): Directory to search for video files
        recursive (bool): Indicate whether to search all subdirectories as well (default = False)
        use_fullpath (bool): Include full filepaths in the returned list (default = False)
        show_size (bool): Prints the disk size of the video files (default = False)
    
    Returns:
        video_files (list): List of names (or full filepaths if `use_fullpath=True`) of the video files

    """

    return listdir(DIR=DIR, recursive=recursive, use_fullpath=use_fullpath, ext = _acceptable_video_formats, verbose=verbose, every=every)


def listdir(
        DIR : str,
        recursive : bool = False, 
        use_fullpath: bool = False, 
        ext : Union[str, List[str]] = None, 
        verbose : (bool, int) = True,
        every : int = 1000
    ) -> List[str]:
    r"""
        Lists all files within a specific directory (and sub-directories if `recursive=True`).
        This can be filtered for certain extensions (by populating `ext`)
    
    Args:
        DIR (str): Directory to search for files
        recursive (bool): Indicate whether to search all subdirectories as well (default = False)
        use_fullpath (bool): Include full filepaths in the returned list (default = False)
        ext (str, list(str), tuple(str)): Filter by extension names.
        show_size (bool): Prints the disk size of the files (default = False)
        verbose (bool): Print info
        every (int): If ``verbose = True``, logging info is displayed after every `every` times.
    
    Returns:
        files (list): List of names (or full filepaths if `use_fullpath=True`) of the files

    """

    if not exists(DIR):
        raise ValueError('Specified directory does not exist')
    
    if not isinstance(recursive, bool):
        raise ValueError('recursive must be a boolean')

    if not isinstance(use_fullpath, bool):
        raise TypeError('use_fullpath must be a boolean')

    if ext is not None:
        if not isinstance(ext, (str, list, tuple)):
            raise TypeError("`ext` must either be of type `str`, or tuple/list of `str`")

        if isinstance(ext, (list, tuple)):
            for i in ext:
                if not isinstance(i, str):
                    raise TypeError("`ext` must be a homogenous list of `str`")
    
    if not isinstance(verbose, bool):
        if isinstance(verbose, int) and verbose not in [0, 1]:
            raise TypeError('verbose must be a boolean (either True or False)')
        raise TypeError('verbose must be a boolean (either True or False)')
    
    if not isinstance(every, int):
        raise TypeError("`every` must be an int")

    dirs : list = []
    count_files : int = 0
    
    start = time()
    count = 0
    if recursive:
        for root, _, files in os.walk(DIR):
            for file in files:
                if ext is not None and not file.endswith(ext):
                    continue
                count += 1
                fullpath = join(root, file).replace('\\', '/')
                if use_fullpath:
                    dirs.append(fullpath)
                else:
                    dirs.append(file)

                if verbose is True and count % every == 0:
                    print(f"[INFO] At {count} files") # come up with a better log message!

    else:
        for file in os.listdir(DIR):
            if ext is not None and not file.endswith(ext):
                continue
            count += 1
            fullpath = join(DIR, file).replace('\\', '/')
            if use_fullpath:
                dirs.append(fullpath)
            else:
                dirs.append(file)
            
            if verbose is True and count % every == 0:
                print(f"[INFO] At {count} files") # come up with a better log message!
    end = time()
    
    if verbose is True:
        count_files = len(dirs)
        if count_files == 1:
            print(f'[INFO] {count_files} file found in {end-start}s')
        else:
            print(f'[INFO] {count_files} files found in {end-start}s')

    return dirs


def is_image(path) -> bool:
    r"""
        Checks if a given path is that of a valid image file
            
    Args:
        path (str): Filepath to check
    
    Returns:
        True; if `path` is a valid image filepath
        False; otherwise

    """

    if not isinstance(path, str):
        raise TypeError('path must be a string')

    if path.endswith(_acceptable_image_formats):
        return True 

    return False


def is_video(path) -> bool:
    r"""
        Checks if a given path is that of a valid image file
            
    Args:
        path (str): Filepath to check
    
    Returns:
        True; if `path` is a valid image filepath
        False; otherwise

    """
    if not isinstance(path, str):
        raise ValueError('path must be a string')

    if path.endswith(_acceptable_video_formats):
        return True 

    return False


def osname() -> str:
    return os.name


def cwd() -> str:
    r"""
        Returns the filepath to the current working directory
    """
    return os.getcwd()


def exists(path) -> bool:
    r"""
        Checks if a given filepath is valid
            
    Args:
        path (str): Filepath to check
    
    Returns:
        True; if `path` is a valid filepath
        False; otherwise

    """

    if not isinstance(path, str):
        raise ValueError('Filepath must be a string')

    return os.path.exists(path)


def isfile(path) -> bool:
    r"""
        Checks if a given filepath is a valid path
            
    Args:
        path (str): Filepath to check
    
    Returns:
        True; if `path` is a valid filepath
        False; otherwise

    """
    return os.path.isfile(path)


def isdir(path) -> bool:
    r"""
        Checks if a given filepath is that of a directory
            
    Args:
        path (str): Filepath to check
    
    Returns:
        True; if `path` is a valid directory
        False; otherwise

    """
    return os.path.isdir(path)


def mkdir(path) -> None:
    r"""
        Creates a directory at `path`

    """
    os.mkdir(path)


def abspath(file_name) -> str:
    r"""
        Returns the absolute path of `file_name`

    """
    return os.path.abspath(file_name)


def chdir(path) -> None:
    r"""
        Checks into directory `path`

    """
    if not isinstance(path, str):
        raise ValueError('Specify a valid path')

    os.chdir(path)


def get_size(file, disp_format='bytes') -> float:
    r"""
        Returns: the size of `file` in bytes/kb/mb/gb/tb
            
    Args:
        file (str): Filepath to check
        disp_format (str): Size format (bytes/kb/mb/gb/tb)
    
    Returns:
        size (str): File size in bytes/kb/mb/gb/tb

    """

    if not isinstance(disp_format, str):
        raise ValueError('display format must be a string')
    
    disp_format = disp_format.lower()

    if disp_format not in ['bytes', 'kb', 'mb', 'gb', 'tb']:
        raise ValueError('display format needs to be either bytes/kb/mb/gb/tb')

    size = os.path.getsize(file)

    if disp_format == 'bytes':
        return size 

    if disp_format == 'kb':
        return size * 1e-3

    if disp_format == 'mb':
        return size * 1e-6

    if disp_format == 'gb':
        return size * 1e-9

    if disp_format == 'tb':
        return size * 1e-12


def join(*paths) -> str:
    r"""
        Join multiple filepaths together

    """

    return os.path.join(*paths)


def dirname(file) -> str:
    r"""
        Returns the base directory name of `file`

    """

    return os.path.dirname(file)



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import numpy as np 

def _check_target_size(size):
    """
    Common check to enforce type and sanity check on size tuples
    :param size: Should be a tuple of size 2 (width, height)
    :returns: True, or raises a ValueError
    """

    if not isinstance(size, (list, tuple)):
        raise ValueError("`target_size` must be a tuple of length 2 `(width,height)`")
    if len(size) != 2:
        raise ValueError("`target_size` must be a tuple of length 2 `(width,height)`")
    if size[0] < 0 or size[1] < 0:
        raise ValueError("Width and height must be >= 0")

    return True


def _check_mean_sub_values(value, channels):
    """
        Checks if mean subtraction values are valid based on the number of channels
        'value' must be a tuple of dimensions = number of channels
    Returns boolean:
        True -> Expression is valid
        False -> Expression is invalid
    """
    if value is None:
        raise ValueError('Value(s) specified is of NoneType()')
    
    if isinstance(value, tuple):
        # If not a tuple, we convert it to one
        try:
            value = tuple(value)
        except TypeError:
            value = tuple([value])
    
    if channels not in [1,3]:
        raise ValueError('Number of channels must be either 1 (Grayscale) or 3 (RGB/BGR)')

    if len(value) not in [1,3]:
        raise ValueError('Tuple length must be either 1 (subtraction over the entire image) or 3 (per channel subtraction)', value)
    
    if len(value) == channels:
        return True 

    else:
        raise ValueError(f'Expected a tuple of dimension {channels}', value) 


def _get_output(array, out, fname, dtype=None, output=None):
    '''
    output = _get_output(array, out, fname, dtype=None, output=None)
    Implements the caer output convention:
        (1) if `out` is None, return np.empty(array.shape, array.dtype)
        (2) else verify that output is of right size, shape, and contiguous
    Parameters
    ----------
    array : Tensor
    out : Tensor or None
    fname : str
        Function name. Used in error messages
    Returns
    -------
    output : Tensor
    '''
    detail = '.\nWhen an output argument is used, the checking is very strict as this is a performance feature.'
    if dtype is None:
        dtype = array.dtype
    if output is not None: # pragma: no cover
        import warnings
        warnings.warn('Using deprecated `output` argument in function `%s`. Please use `out` in the future. It has exactly the same meaning and it matches what numpy uses.' % fname, DeprecationWarning)
        if out is not None:
            warnings.warn('Using both `out` and `output` in function `%s`.\ncaer is going to ignore the `output` argument and use the `out` version exclusively.' % fname)
        else:
            out = output

    if out is None:
        return np.empty(array.shape, dtype)

    if out.dtype != dtype:
        raise ValueError(
            'caer.%s: `out` has wrong type (out.dtype is %s; expected %s)%s' %
                (fname, out.dtype, dtype, detail))

    if out.shape != array.shape:
        raise ValueError('caer.%s: `out` has wrong shape (got %s, while expecting %s)%s' % (fname, out.shape, array.shape, detail))

    if not out.flags.contiguous:
        raise ValueError('caer.%s: `out` is not c-array%s' % (fname,detail))

    return out

def _get_axis(array, axis, fname):
    '''
    axis = _get_axis(array, axis, fname)
    Checks that ``axis`` is a valid axis of ``array`` and normalises it.
    Parameters
    ----------
    array : Tensor
    axis : int
    fname : str
        Function name. Used in error messages
    Returns
    -------
    axis : int
        The positive index of the axis to use
    '''
    if axis < 0:
        axis += len(array.shape)

    if not (0 <= axis < len(array.shape)):
        raise ValueError('caer.%s: `axis` is out of bounds (maximum was %s, got %s)' % (fname, array.ndim, axis))

    return axis


def _normalize_sequence(array, value, fname):
    '''
    values = _normalize_sequence(array, value, fname)
    If `value` is a sequence, checks that it has an element for each dimension
    of `array`. Otherwise, returns a sequence that repeats `value` once for
    each dimension of array.
    Parameters
    ----------
    array : Tensor
    value : sequence or scalar
    fname : str
        Function name. Used in error messages
    Returns
    -------
    values : sequence
    '''
    try:
        value = list(value)
    except TypeError:
        return [value for s in array.shape]
    if len(value) != array.ndim:
        raise ValueError('caer.%s: argument is sequence, but has wrong size (%s for an array of %s dimensions)' % (fname, len(value), array.ndim))
    return value


def _verify_is_floatingpoint_type(A, function_name):
    '''
    _verify_is_integer_type(array, "function")
    Checks that ``A`` is a floating-point array. If it is not, it raises
    ``TypeError``.
    Parameters
    ----------
    A : Tensor
    function_name : str
        Used for error messages
    '''
    if not np.issubdtype(A.dtype, np.floating):
        raise TypeError('caer.{}: This function only accepts floating-point types (passed array of type {})'.format(function_name, A.dtype))


def _verify_is_integer_type(A, function_name):
    '''
    _verify_is_integer_type(array, "function")
    Checks that ``A`` is an integer array. If it is not, it raises
    ``TypeError``.
    Parameters
    ----------
    A : Tensor
    function_name : str
        Used for error messages
    '''
    k = A.dtype.kind
    if k not in "iub": # integer, unsigned integer, boolean
        raise TypeError('caer.%s: This function only accepts integer types (passed array of type %s)' % (function_name, A.dtype))


def _verify_is_nonnegative_integer_type(A, function_name):
    '''
    _verify_is_nonnegative_integer_type(array, "function")
    Checks that ``A`` is an unsigned integer array. If it is not, it raises
    ``TypeError``.
    Parameters
    ----------
    A : Tensor
    function_name : str
        Used for error messages
    '''
    _verify_is_integer_type(A, function_name)
    if A.dtype.kind == 'i' and not np.all(A >= 0):
        raise ValueError('caer.{0}: This function only accepts positive integer types (passed array of type {1})'.format(function_name, A.dtype))


def _make_binary(array):
    '''
    bin = _make_binary(array)
    Returns (possibly a copy) of array as a boolean array
    '''
    array = np.asanyarray(array)
    if array.dtype != bool:
        return (array != 0)
    return array


def _as_floating_point_array(array):
    '''
    array = _as_floating_point_array(array)
    Returns (possibly a copy) of array as a floating-point array
    '''
    array = np.asanyarray(array)
    if not np.issubdtype(array.dtype, np.floating):
        return array.astype(np.double)
    return array


def _check_3(arr, funcname):
    if arr.ndim != 3 or arr.shape[2] != 3:
        raise ValueError('caer.%s: this function expects an array of shape (h, w, 3), received an array of shape %s.' % (funcname, arr.shape))


def _check_2(arr, funcname):
    if arr.ndim != 2:
        raise ValueError('caer.%s: this function can only handle 2D arrays (passed array with shape %s).' % (funcname, arr.shape))


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

#pylint:disable=unused-argument,unused-variable,eval-used

import numpy as np 
import random
import cv2 as cv 
import math

from ..coreten import Tensor, to_tensor

from .functional import (
    is_list,
    _hls,
    _exposure_process
)


__all__ = [
    'adjust_brightness',
    'adjust_contrast',
    'adjust_hue',
    'adjust_saturation',
    'adjust_gamma',
    'affine',
    'darken',
    'brighten',
    'random_brightness',
    'correct_exposure',
    'augment_random'
]


def adjust_brightness(tens, coeff, rgb=True) -> Tensor:
    r"""
        Adjust the brightness of an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        coeff (int): Coefficient value.
            - ``coeff < 1``, the image is darkened.
            - ``coeff = 1``, the image is unchanged.
            - ``coeff > 1``, the image is lightened.
        rgb (bool): Operate on RGB images. Default: True.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise(rgb=True)
        >> filtered = caer.transforms.adjust_brightness(tens, coeff=1.4, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    tens = to_tensor(tens, enforce_tensor=True)
    tens = _hls(tens)
    cspace = tens.cspace 

    tens = np.array(tens, dtype=np.float64) 
    tens[:,:,1] = tens[:,:,1] * coeff ## scale pixel values up or down for channel 1 (for lightness)

    if coeff > 1:
        tens[:,:,1][tens[:,:,1]>255]  = 255 # Set all values > 255 to 255
    else:
        tens[:,:,1][tens[:,:,1]<0] = 0

    return to_tensor(tens, cspace=cspace, dtype=np.uint8)


def brighten(tens, coeff=-1, rgb=True) -> Tensor:
    r"""
        Brighten an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        coeff (int): Coefficient value.
        rgb (bool): Operate on RGB images. Default: True.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise(rgb=True)
        >> filtered = caer.transforms.brighten(tens, coeff=-1, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace

    if coeff !=-1:
        if coeff < 0.0 or coeff > 1.0:
            raise ValueError('Brightness coefficient can only be between 0.0 and 1.0')

    if coeff == -1:
        coeff_t = 1 + random.uniform(0, 1) # coeff between 1.0 and 1.5
    else:
        coeff_t = 1 + coeff  # coeff between 1.0 and 2.0

    tens = adjust_brightness(tens, coeff_t)
    return to_tensor(tens, cspace=cspace)


def darken(tens, darkness_coeff = -1) -> Tensor:
    r"""
        Darken an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        darkness_coeff (int): Coefficient value.
        rgb (bool): Operate on RGB images. Default: True.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise(rgb=True)
        >> filtered = caer.transforms.darken(tens, coeff=-1, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace 

    if darkness_coeff != -1:
        if darkness_coeff < 0.0 or darkness_coeff > 1.0:
            raise ValueError('Darkness coeff must only be between 0.0 and 1.0') 

    if darkness_coeff == -1:
        darkness_coeff_t = 1 - random.uniform(0, 1)
    else:
        darkness_coeff_t = 1 - darkness_coeff  

    tens = adjust_brightness(tens, darkness_coeff_t)
    return to_tensor(tens, cspace=cspace)


def random_brightness(tens) -> Tensor:
    r"""
        Add random brightness to an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise()
        >> filtered = caer.transforms.random_brightness(tens, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    rand_br_coeff = 2 * np.random.uniform(0, 1) # Generates a value between 0.0 and 2.0
    return adjust_brightness(tens, rand_br_coeff)


def adjust_contrast(tens, contrast_factor) -> Tensor:
    """
        Adjust contrast of an image.

    Args:
        tens (Tensor): Any valid ``caer.Tensor``.
        contrast_factor (float): How much to adjust the contrast. Can be any
            non negative number. 0 gives a solid gray image, 1 gives the
            original image while 2 increases the contrast by a factor of 2.
    Returns:
        ``caer.Tensor``: Contrast adjusted image.
    """
    # It's much faster to use the LUT construction because you have to change dtypes multiple times
    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace 

    table = np.array([(i - 74) * contrast_factor + 74
                      for i in range(0, 256)]).clip(0, 255).astype('uint8')

    try:
        from PIL import ImageEnhance
    except ImportError:
        raise ImportError('Pillow must be installed to use ``caer.color.adjust_saturation()``.')

    enhancer = ImageEnhance.Contrast(tens)
    tens = enhancer.enhance(contrast_factor)
    tens = cv.LUT(tens, table)
    return to_tensor(tens, cspace=cspace)


def adjust_saturation(tens, saturation_factor) -> Tensor:
    """Adjust color saturation of an image.

    Args:
        tens (Tensor): Any valid ``caer.Tensor``.
        saturation_factor (float):  How much to adjust the saturation. 0 will
            give a black and white image, 1 will give the original image while
            2 will enhance the saturation by a factor of 2.

    Returns:
        Saturation-adjusted image.
    
    Examples::

        >> tens = caer.data.sunrise(rgb=True)
        >> filtered = caer.transforms.adjust_saturation(tens, saturation_factor=1.5, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace 

    try:
        from PIL import Image, ImageEnhance
    except ImportError:
        raise ImportError('Pillow must be installed to use ``caer.color.adjust_saturation()``.')

    tens = Image.fromarray(tens)
    enhancer = ImageEnhance.Color(tens)
    tens = enhancer.enhance(saturation_factor)

    return to_tensor(tens, cspace=cspace)


def adjust_hue(tens, hue_factor) -> Tensor:
    r"""
        Adjust hue of an image.

        The image hue is adjusted by converting the image to HSV and cyclically shifting the intensities in the hue channel (H). The image is then converted back to original image mode.
        
        See `Hue`_ for more details.
        .. _Hue: https://en.wikipedia.org/wiki/Hue

    Args:
        tens (Tensor): Any valid ``caer.Tensor``.
        hue_factor (float):  How much to shift the hue channel. Should be in the range [-0.5, 0.5]. 
            0.5 and -0.5 give complete reversal of hue channel in HSV space in positive and negative direction respectively.
            0 means no shift. Therefore, both -0.5 and 0.5 will give an image with complementary colors while 0 gives the original image.
            
    Returns:
        Hue adjusted image.

    Examples::

        >> tens = caer.data.sunrise(rgb=True)
        >> filtered = caer.transforms.adjust_hue(tens, hue_factor=1, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace 

    if not (-0.5 <= hue_factor <= 0.5):
        raise ValueError('`hue_factor` is not in [-0.5, 0.5].')

    try:
        from PIL import Image
    except ImportError:
        raise ImportError('Pillow must be installed to use this ``caer.color.adjust_hue()``.')

    tens = Image.fromarray(tens)
    input_mode = tens.mode
    if input_mode in {'L', '1', 'I', 'F'}:
        return np.array(tens)

    h, s, v = tens.convert('HSV').split()

    np_h = np.array(h, dtype=np.uint8)

    # uint8 addition take cares of rotation across boundaries
    with np.errstate(over='ignore'):
        np_h += np.uint8(hue_factor * 255)
    h = Image.fromarray(np_h, 'L')

    tens = Image.merge('HSV', (h, s, v)).convert(input_mode)

    return to_tensor(tens, cspace=cspace)


def adjust_gamma(tens, gamma, gain=1) -> Tensor:
    r"""
        Perform gamma correction on an image.

        Also known as Power Law Transform. Intensities in RGB mode are adjusted
        based on the following equation:
        .. math::
            I_{\text{out}} = 255 \times \text{gain} \times \left(\frac{I_{\text{in}}}{255}\right)^{\gamma}
        See `Gamma Correction`_ for more details.
        .. _Gamma Correction: https://en.wikipedia.org/wiki/Gamma_correction

    Args:
        tens (Tensor): Any valid ``caer.Tensor``.
        gamma (float): Non negative real number, same as :math:`\gamma` in the equation.
            gamma larger than 1 make the shadows darker,
            while gamma smaller than 1 make dark regions lighter.
        gain (float): The constant multiplier.

    Examples::

        >> tens = caer.data.sunrise(rgb=True)
        >> filtered = caer.transforms.adjust_gamma(tens, gamma=1.5, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace 

    if gamma < 0:
        raise ValueError('Gamma should be a non-negative real number')

    # from here
    # https://stackoverflow.com/questions/33322488/how-to-change-image-illumination-in-opencv-python/41061351
    table = np.array([((i / 255.0)**gamma) * 255 * gain
                      for i in np.arange(0, 256)]).astype('uint8')

    tens = cv.LUT(tens, table)
    return to_tensor(tens, cspace=cspace)


def _get_affine_matrix(center, angle, translate, scale, shear) -> Tensor:
    # Helper method to compute matrix for affine transformation
    # We need compute affine transformation matrix: M = T * C * RSS * C^-1
    # where T is translation matrix: [1, 0, tx | 0, 1, ty | 0, 0, 1]
    #       C is translation matrix to keep center: [1, 0, cx | 0, 1, cy | 0, 0, 1]
    #       RSS is rotation with scale and shear matrix
    #       RSS(a, scale, shear) = [ cos(a)*scale    -sin(a + shear)*scale     0]
    #                              [ sin(a)*scale    cos(a + shear)*scale     0]
    #                              [     0                  0          1]

    angle = math.radians(angle)
    shear = math.radians(shear)
    # scale = 1.0 / scale

    T = np.array([[1, 0, translate[0]], [0, 1, translate[1]], [0, 0, 1]])
    C = np.array([[1, 0, center[0]], [0, 1, center[1]], [0, 0, 1]])
    RSS = np.array(
        [[math.cos(angle) * scale, -math.sin(angle + shear) * scale, 0],
         [math.sin(angle) * scale,
          math.cos(angle + shear) * scale, 0], [0, 0, 1]])
    matrix = T @ C @ RSS @ np.linalg.inv(C)

    return matrix[:2, :]


def affine(tens, angle, translate, scale, shear, interpolation='bilinear', mode=0, fillcolor=0) -> Tensor:
    """
        Apply affine transformation on the image keeping image center invariant.

    Args:
        tens (Tensor): Any valid ``caer.Tensor``.
        angle (float or int): Rotation angle in degrees between -180 and 180, clockwise direction.
        translate (list or tuple of integers): Horizontal and vertical translations (post-rotation translation)
        scale (float): Overall scale
        shear (float): Shear angle value in degrees between -180 to 180, clockwise direction.
        interpolation (int, str): Interpolation to use for resizing. Defaults to `'bilinear'`. 
                Supports `'bilinear'`, `'bicubic'`, `'area'`, `'nearest'`.
        mode (int, str): Method for filling in border regions. 
                Defaults to ``constant`` meaning areas outside the image are filled with a value (val, default 0). 
                Supports ``'replicate'``, ``'reflect'``, ``'reflect-101'``.
        val (int): Optional fill color for the area outside the transform in the output image. Default: 0
    """

    tens = to_tensor(tens, enforce_tensor=True)
    cspace = tens.cspace 

    assert isinstance(translate, (tuple, list)) and len(translate) == 2, \
        'Argument translate should be a list or tuple of length 2'

    assert scale > 0.0, 'Argument scale should be positive'

    interpolation_methods = {
        'nearest': 0,  '0': 0,  0: 0, # 0
        'bilinear': 1, '1': 1,  1: 1, # 1
        'bicubic': 2,  '2': 2,  2: 2, # 2
        'area': 3,     '3': 3,  3: 3 # 3
    }
    border_methods = {
        'constant': 0,    '0': 0, 0: 0, # 0
        'replicate': 1,   '1': 1, 1: 1, # 1
        'reflect': 2,     '2': 2, 2: 2, # 2
        'reflect-101': 4, '4': 4, 4: 4 # 4
    }

    if interpolation not in interpolation_methods:
        raise ValueError('Specify a valid interpolation type - area/nearest/bicubic/bilinear')

    if mode not in border_methods:
        raise ValueError('Specify a valid border type - constant/replicate/reflect/reflect-101')

    output_size = tens.shape[0:2]
    center = (tens.shape[1] * 0.5 + 0.5, tens.shape[0] * 0.5 + 0.5)
    matrix = _get_affine_matrix(center, angle, translate, scale, shear)

    tens = cv.warpAffine(tens,
                            matrix,
                            output_size[::-1],
                            interpolation,
                            borderMode=mode,
                            borderValue=fillcolor)

    return to_tensor(tens, cspace=cspace)


def correct_exposure(tens, rgb=True) -> Tensor:
    r"""
        Correct the exposure of an image.

    Args:
        tens (Tensor) : Any regular ``caer.Tensor``.
        rgb (bool): Operate on RGB images. Default: True.
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise(rgb=True)
        >> filtered = caer.transforms.correct_exposure(tens, rgb=True)
        >> filtered
        (427, 640, 3)

    """
    return _exposure_process(tens)


def augment_random(tens, aug_types='', volume='expand' ) -> Tensor:
    aug_types_all = ['random_brightness','add_shadow','add_snow','add_rain','add_fog','add_gravel','add_sun_flare','add_motion_blur','add_autumn','random_flip']

    if aug_types=='':
        aug_types=aug_types_all
    output=[]

    if not is_list(aug_types):
        raise ValueError('`aug_types` should be a list of function names (str)')
    
    if volume == 'expand':
        for aug_type in aug_types:
            if not(aug_type in aug_types_all):
                raise ValueError('Incorrect transformation function defined')

            command = aug_type + '(tens)'
            result = eval(command)

            output.append(result)

    elif volume == 'same':
        for aug_type in aug_types:
            if not (aug_type in aug_types_all):
                raise ValueError('Incorrect transformation function defined')

        selected_aug = aug_types[random.randint(0, len(aug_types)-1)]
        command = selected_aug+'(tens)'
        output = eval(command)

    else: 
        raise ValueError('volume type can only be "same" or "expand"')

    return output



#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

from ..coreten import Tensor, to_tensor

from ._bgr import bgr2gray, bgr2hsv, bgr2lab, bgr2rgb, bgr2hls, bgr2yuv, bgr2luv
from ._rgb import rgb2gray, rgb2hsv, rgb2lab, rgb2bgr, rgb2hls, rgb2yuv, rgb2luv
from ._gray import gray2lab, gray2rgb, gray2hsv, gray2bgr, gray2hls, gray2yuv, gray2luv
from ._hsv import hsv2gray, hsv2rgb, hsv2lab, hsv2bgr, hsv2hls, hsv2yuv, hsv2luv
from ._hls import hls2gray, hls2rgb, hls2lab, hls2bgr, hls2hsv, hls2yuv, hls2luv
from ._lab import lab2gray, lab2rgb, lab2hsv, lab2bgr, lab2hls, lab2yuv, lab2luv
from ._yuv import yuv2bgr, yuv2rgb, yuv2hsv, yuv2hls, yuv2gray, yuv2lab, yuv2luv
from ._luv import luv2bgr, luv2rgb, luv2hsv, luv2hls, luv2gray, luv2lab, luv2yuv

__all__ = [
    'to_rgb',
    'to_bgr',
    'to_gray',
    'to_hsv',
    'to_hls',
    'to_lab',
    'to_yuv',
    'to_luv',
]


def to_rgb(tens) -> Tensor:
    r'''
        Converts any supported colorspace to RGB

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    # If 'null', we assume we have a brand new Tensor
    if cspace == 'null':
        print(
            'Warning: Caer was unable to assign a colorspace for a foreign tensor. Sticking with "rgb". This issue will be'
            'fixed in a future update.'
        )

        # We assume that the tens is a BGR image
        im = bgr2rgb(tens)
        im = to_tensor(im, cspace='rgb')
        return im

    elif cspace == 'rgb':
        return tens

    elif cspace == 'bgr':
        im = bgr2rgb(tens)
        return to_tensor(im, cspace='rgb')

    elif cspace == 'gray':
        im = gray2rgb(tens)
        return to_tensor(im, cspace='rgb')

    elif cspace == 'hls':
        im = hls2rgb(tens)
        return to_tensor(im, cspace='rgb')

    elif cspace == 'hsv':
        im = hsv2rgb(tens)
        return to_tensor(im, cspace='rgb')

    elif cspace == 'lab':
        im = lab2rgb(tens)
        return to_tensor(im, cspace='rgb')

    elif cspace == 'yuv':
        im = yuv2rgb(tens)
        return to_tensor(im, cspace='rgb')

    elif cspace == 'luv':
        im = luv2rgb(tens)
        return to_tensor(im, cspace='rgb')


def to_bgr(tens) -> Tensor:
    r'''
        Converts any supported colorspace to BGR

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    if cspace == 'bgr':
        return tens

    elif cspace == 'gray':
        im = gray2bgr(tens)
        return to_tensor(im, cspace='bgr')

    elif cspace == 'rgb':
        im = rgb2bgr(tens)
        return to_tensor(im, cspace='bgr')

    elif cspace == 'hls':
        im = hls2bgr(tens)
        return to_tensor(im, cspace='bgr')

    elif cspace == 'hsv':
        im = hsv2bgr(tens)
        return to_tensor(im, cspace='bgr')

    elif cspace == 'lab':
        im = lab2bgr(tens)
        return to_tensor(im, cspace='bgr')

    elif cspace == 'yuv':
        im = yuv2bgr(tens)
        return to_tensor(im, cspace='bgr')

    elif cspace == 'luv':
        im = luv2bgr(tens)
        return to_tensor(im, cspace='bgr')


def to_gray(tens) -> Tensor:
    r'''
        Converts any supported colorspace to grayscale

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    if cspace == 'gray':
        return tens

    elif cspace == 'bgr':
        im = bgr2gray(tens)
        return to_tensor(im, cspace='gray')

    elif cspace == 'rgb':
        im = rgb2gray(tens)
        return to_tensor(im, cspace='gray')

    elif cspace == 'hls':
        im = hls2gray(tens)
        return to_tensor(im, cspace='gray')

    elif cspace == 'hsv':
        im = hsv2gray(tens)
        return to_tensor(im, cspace='gray')

    elif cspace == 'lab':
        im = lab2gray(tens)
        return to_tensor(im, cspace='gray')

    elif cspace == 'yuv':
        im = yuv2gray(tens)
        return to_tensor(im, cspace='gray')

    elif cspace == 'luv':
        im = luv2gray(tens)
        return to_tensor(im, cspace='gray')


def to_hsv(tens) -> Tensor:
    r'''
        Converts any supported colorspace to hsv

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    if cspace == 'hsv':
        return tens

    elif cspace == 'bgr':
        im = bgr2hsv(tens)
        return to_tensor(im, cspace='hsv')

    elif cspace == 'rgb':
        im = rgb2hsv(tens)
        return to_tensor(im, cspace='hsv')

    elif cspace == 'hls':
        im = hls2hsv(tens)
        return to_tensor(im, cspace='hsv')

    elif cspace == 'gray':
        im = gray2hsv(tens)
        return to_tensor(im, cspace='hsv')

    elif cspace == 'lab':
        im = lab2hsv(tens)
        return to_tensor(im, cspace='hsv')

    elif cspace == 'yuv':
        im = yuv2hsv(tens)
        return to_tensor(im, cspace='hsv')

    elif cspace == 'luv':
        im = luv2hsv(tens)
        return to_tensor(im, cspace='hsv')


def to_hls(tens) -> Tensor:
    r'''
        Converts any supported colorspace to HLS

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    if cspace == 'hls':
        return tens

    elif cspace == 'bgr':
        im = bgr2hls(tens)
        return to_tensor(im, cspace='hls')

    elif cspace == 'rgb':
        im = rgb2hls(tens)
        return to_tensor(im, cspace='hls')

    elif cspace == 'hsv':
        im = hsv2hls(tens)
        return to_tensor(im, cspace='hls')

    elif cspace == 'gray':
        im = gray2hls(tens)
        return to_tensor(im, cspace='hls')

    elif cspace == 'lab':
        im = lab2hls(tens)
        return to_tensor(im, cspace='hls')

    elif cspace == 'yuv':
        im = yuv2hls(tens)
        return to_tensor(im, cspace='hls')

    elif cspace == 'luv':
        im = luv2hls(tens)
        return to_tensor(im, cspace='hls')


def to_lab(tens) -> Tensor:
    r'''
        Converts any supported colorspace to LAB

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    if cspace == 'lab':
        return tens

    elif cspace == 'bgr':
        im = bgr2lab(tens)
        return to_tensor(im, cspace='lab')
    elif cspace == 'rgb':
        im = rgb2lab(tens)
        return to_tensor(im, cspace='lab')
    elif cspace == 'hsv':
        im = hsv2lab(tens)
        return to_tensor(im, cspace='lab')
    elif cspace == 'gray':
        im = gray2lab(tens)
        return to_tensor(im, cspace='lab')
    elif cspace == 'hls':
        im = hls2lab(tens)
        return to_tensor(im, cspace='lab')
    elif cspace == 'yuv':
        im = yuv2lab(tens)
        return to_tensor(im, cspace='lab')
    elif cspace == 'luv':
        im = luv2lab(tens)
        return to_tensor(im, cspace='lab')


def to_yuv(tens) -> Tensor:
    r'''
        Converts any supported colorspace to YUV

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    if cspace == 'yuv':
        return tens

    elif cspace == 'bgr':
        im = bgr2yuv(tens)
        return to_tensor(im, cspace='yuv')
    elif cspace == 'rgb':
        im = rgb2yuv(tens)
        return to_tensor(im, cspace='yuv')
    elif cspace == 'hsv':
        im = hsv2yuv(tens)
        return to_tensor(im, cspace='yuv')
    elif cspace == 'gray':
        im = gray2yuv(tens)
        return to_tensor(im, cspace='yuv')
    elif cspace == 'hls':
        im = hls2yuv(tens)
        return to_tensor(im, cspace='yuv')
    elif cspace == 'lab':
        im = lab2yuv(tens)
        return to_tensor(im, cspace='yuv')
    elif cspace == 'luv':
        im = luv2yuv(tens)
        return to_tensor(im, cspace='yuv')


def to_luv(tens) -> Tensor:
    r'''
        Converts any supported colorspace to LUV

    Args:
        tens (Tensor)

    Returns:
        Tensor
    '''
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor
    _ = (
        tens._nullprt()
    )  # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value
    cspace = tens.cspace

    if cspace == 'luv':
        return tens

    elif cspace == 'bgr':
        im = bgr2luv(tens)
        return to_tensor(im, cspace='luv')
    elif cspace == 'rgb':
        im = rgb2luv(tens)
        return to_tensor(im, cspace='luv')
    elif cspace == 'hsv':
        im = hsv2luv(tens)
        return to_tensor(im, cspace='luv')
    elif cspace == 'gray':
        im = gray2luv(tens)
        return to_tensor(im, cspace='luv')
    elif cspace == 'hls':
        im = hls2luv(tens)
        return to_tensor(im, cspace='luv')
    elif cspace == 'lab':
        im = lab2luv(tens)
        return to_tensor(im, cspace='luv')
    elif cspace == 'yuv':
        im = yuv2luv(tens)
        return to_tensor(im, cspace='luv')



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


#pylint:disable=no-member,pointless-string-statement

from threading import Thread
import math
import cv2 as cv

from .constants import FPS

__all__ = [
    'LiveStream'
]

"""
    Python threading has a specific meaning for daemon. A daemon thread will shut down immediately when the program exits. One way to think about these definitions is to consider the daemon thread a thread that runs in the background without worrying about shutting it down.

    If a program is running Threads that are not daemons, then the program will wait for those threads to complete before it terminates. Threads that are daemons, however, are just killed wherever they are when the program is exiting.
"""

# This class can only handle live video streams. When applied on pre-existing videos, there appears to # be an issue with Threading. As a result, the video plays through with a high (almost x4) speed
# This issue has been marked and will be fixed in a future update. 

class LiveStream:
    r"""
        This is an auxiliary class that enables Live Video Streaming for caer with minimalistic latency, and at the expense
        of little to no additional computational requirements.
        
        The basic idea behind it is to tracks and save the salient feature array for the given number of frames and then uses these anchor point to cancel out all perturbations relative to it for the incoming frames in the queue. This class relies heavily on **Threaded Queue mode** for error-free & ultra-fast frame handling.

    Args:
        source (int): Source path for the video. If ``source=0``, the default camera device is used. For 
            multiple external camera devices, use incremented values. For eg: ``source=1`` represents the second camera device on your system.
    """

    def __init__(self, source=0):
        r"""
            Source must either be an integer (0, 1, 2 etc) or a path to a video file
        """
        if isinstance(source, str):
            raise ValueError('Expected an integer. Got a filepath. LiveVideoStream is for live streams only')
        
        # Initializing the video stream
        self.stream = cv.VideoCapture(source)
        print('Live yes')
        self.ret, self.frame = self.stream.read()
        print('Live not')

        self.width = int(self.stream.get(cv.CAP_PROP_FRAME_WIDTH))
        self.height = int(self.stream.get(cv.CAP_PROP_FRAME_HEIGHT))
        self.res = (self.width, self.height)

        self.fps = math.ceil(self.stream.get(FPS))

        # Initializing the thread name
        self.thread_name = 'DefaultVideoStream'

        # Boolean to check whether stream should be killed
        self.kill_stream = False

    def begin_stream(self):
        # Starting the thread to read frames from the video stream
        thread = Thread(target=self.update, name=self.thread_name, args=())
        thread.daemon = True
        thread.start()
        return self

    def read(self):
        return self.frame
    
    def update(self):
        while not self.kill_stream:
            self.ret, self.frame = self.stream.read()
    
    def release(self):
        # Stops the stream
        # Releases video pointer
        self.kill_stream = True
    
    # Counting frames not applicable for live video
    
    # # Gets frame count
    # def count_frames(self):
    #     if not self.kill_stream:
    #         if get_opencv_version() == '2':
    #             return int(self.stream.get(FRAME_COUNT_DEPR))
    #         else:
    #             return int(self.stream.get(FRAME_COUNT))

    # Gets FPS count
    def get_fps(self):
        if not self.kill_stream:
            return self.fps
            # if get_opencv_version() == '2':
            #     return math.ceil(self.stream.get(FPS_DEPR))
            # else:
            #     return math.ceil(self.stream.get(FPS))

    
    # Get frame dimensions
    def get_res(self):
        return self.res


r""" \
Caer
=====
Caer is a *lightweight, scalable* Computer Vision library for high-performance AI research. It simplifies your approach towards Computer Vision by abstracting away unnecessary boilerplate code giving you the flexibility to quickly prototype deep learning models or research ideas. 

Our design philosophy makes Caer ideal for students, researchers, hobbyists and even experts in the fields of Deep Learning and Computer Vision.

Documentation: https://caer.rtfd.io

Available subpackages
---------------------
color
    Colorspace manipulation
data
    Standard high-quality test images
io
    Input/Output
path
    OS-specific Path Manipulations
transforms
    Image transformations & augmentations
video
    Video processing utilities


Utilities
---------
__version__
    Caer version string

__author__
    Author of Caer

__contributors__
    List of all contributors to the project

__homepage__
    Web URL of the Caer documentation

"""
#pylint:disable=broad-except

import sys 
import platform

MAJOR = 2
MINOR = 0
MICRO = 6
ISRELEASED = True
VERSION = f'{MAJOR}.{MINOR}.{MICRO}'

RUN_CYTHON_BUILD = False

min_version = (3, 6, 1)

def is_right_py_version(min_py_version):
    if sys.version_info < (3,):
        sys.stderr.write('Python 2 has reached end-of-life and is no longer supported by Caer.')
        return False

    if sys.version_info < min_py_version:
        python_min_version_str = '.'.join((str(num) for num in min_py_version))
        no_go = f'You are using Python {platform.python_version()}. Python >={python_min_version_str} is  required.'
        sys.stderr.write(no_go)
        return False

    return True

if not is_right_py_version(min_version):
    sys.exit(-1)


from setuptools import setup, find_packages
# from setuptools import Extension
from distutils.command.build_ext import build_ext
from configparser import ConfigParser
import subprocess
import os 

# Git version stuff
sha = 'Unknown'
here = os.path.dirname(os.path.abspath(__file__))

try:
    sha = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=here).decode('ascii').strip()
except Exception:
    pass


# Configurations

# All settings are in configs.ini
config = ConfigParser(delimiters=['='])
config.read('configs.ini')
cfg = config['metadata']

cfg_keys = 'description keywords author author_email contributors'.split()
expected = cfg_keys + "name user git_branch license status audience language dev_language".split()
for i in expected: assert i in cfg, f'Missing expected setting: {i}'

# def set_builtin(name, value):
#     if isinstance(__builtins__, dict):
#         __builtins__[name] = value
#     else:
#         setattr(__builtins__, name, value)

# # Prevent numpy from thinking it is still in its setup process:
# set_builtin('__NUMPY_SETUP__', False)
# import numpy as np 


# Defining Setup Variables

NAME = cfg['name']
AUTHOR = cfg['author']
AUTHOR_EMAIL = cfg['author_email']
AUTHOR_LONG = AUTHOR + ' <' + AUTHOR_EMAIL + '>'
LICENSE = cfg['license']
FUNDING = cfg['funding']
GIT_VERSION = repr(sha)
PLATFORMS = ['Any']
GIT_URL = cfg['git_url']
DOCS_URL = cfg['docs_url']
DOWNLOAD_URL = cfg['download_url']
PACKAGES = [i for i in find_packages() if 'tests' not in i]
DESCRIPTION = cfg['description']
LONG_DESCRIPTION = open('README.md', encoding='utf-8').read()
KEYWORDS = [i for i in cfg['keywords'].split(', ')]
REQUIREMENTS = [i for i in cfg['pip_requirements'].split(', ')]
CLASSIFIERS = [i for i in cfg['classifiers'].split('\n')][1:]
PYTHON_REQUIRES = '>=' + cfg['min_python']
# EXTENSIONS = {
#     'caer.filters.cconvex': ['caer/filters/cconvex.cpp'],
#     'caer.filters.cconvolve': ['caer/filters/cconvolve.cpp', 'caer/src/cfilters.cpp'],
#     'caer.distance.cdistance': ['caer/distance/cdistance.cpp'],
#     'caer.morph.cmorph': ['caer/morph/cmorph.cpp', 'caer/src/cfilters.cpp'],
#     'caer.ndi.cndi' : ['caer/ndi/include/cndimage.c', 
#                        'caer/ndi/include/cndfilters.c',
#                        'caer/ndi/include/cndfourier.c',
#                        'caer/ndi/include/cndinterpolation.c',
#                        'caer/ndi/include/cndmeasure.c',
#                        'caer/ndi/include/cndmorphology.c',
#                        'caer/ndi/include/cndsplines.c',
#                        'caer/ndi/include/cndsupport.c'
#                 ]
# }
# EXT_MODULES = [Extension(key, sources=sources, include_dirs=[np.get_include()]) for key, sources in EXTENSIONS.items()]

STATUSES = [ 
    '1 - Planning', 
    '2 - Pre-Alpha', 
    '3 - Alpha',
    '4 - Beta', 
    '5 - Production/Stable', 
    '6 - Mature', 
    '7 - Inactive' 
]


META_PY_TEXT =\
"""
# This file is automatically generated during the generation of setup.py
# Copyright 2020-2021, Caer
author = '%(author)s'
author_email = '%(author_email)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_version = %(git_version)s
release = %(isrelease)s
contributors = %(contributors)s
homepage = '%(homepage)s'
"""


def get_contributors_list(filename='CONTRIBUTORS'):
    contr = [] 
    with open(filename, 'r') as a:
        for line in a:
            line = line.strip()
            # line = """ + line + """
            contr.append(line)
    return contr


def write_meta(filename='caer/_meta.py'):
    print('[INFO] Writing _meta.py')
    TEXT = META_PY_TEXT
    FULL_VERSION = VERSION
    CONTRIBUTORS = get_contributors_list()
    HOMEPAGE = GIT_URL
    
    a = open(filename, 'w')

    try:
        a.write(TEXT % {'author': AUTHOR,
                        'author_email': str(AUTHOR_EMAIL),
                        'version': VERSION,
                       'full_version': FULL_VERSION,
                       'git_version': GIT_VERSION,
                       'isrelease': str(ISRELEASED),
                       'contributors': CONTRIBUTORS,
                       'homepage': str(HOMEPAGE) })
    finally:
        a.close()


def get_docs_url():
    return DOCS_URL


CYTHON_SOURCES = ('',)
def generate_cython():
    cwd = os.path.abspath(os.path.dirname(__file__))
    print('[INFO] Cythonizing sources')
    p = subprocess.call([sys.executable,
                        os.path.join(cwd, 'tools', 'cythonize.py')],
                        cwd=cwd)
    if p != 0:
        raise RuntimeError('[ERROR] Running cythonize failed!')


copt={
    'msvc': ['/EHsc'], 
    'intelw': ['/EHsc']  
}

class build_extension_class(build_ext):
    def build_extensions(self):
        c = self.compiler.compiler_type
        if c in copt:
            for e in self.extensions:
                e.extra_compile_args = copt[c]
        build_ext.build_extensions(self)

CMDCLASS = {
    'build_ext': build_extension_class
}


def setup_package():
    # Rewrite the meta file everytime
    write_meta()

    metadata = dict(
        name = NAME,
        version = VERSION,
        author = AUTHOR,
        author_email = AUTHOR_EMAIL,
        maintainer = AUTHOR,
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        long_description_content_type = 'text/markdown',
        url = GIT_URL,
        download_url = DOWNLOAD_URL,
        project_urls = {
            'Bug Tracker': GIT_URL + '/issues',
            'Documentation': get_docs_url(),
            'Funding': FUNDING,
            'Source Code': GIT_URL,
        },
        packages = PACKAGES,
        # ext_modules = EXT_MODULES,
        license = LICENSE,
        platforms = PLATFORMS,
        install_requires = REQUIREMENTS,
        python_requires = PYTHON_REQUIRES,
        # cmdclass = CMDCLASS,
# Include_package_data is required for setup.py to recognize the MAINFEST.in file
# https://python-packaging.readthedocs.io/en/latest/non-code-files.html
        include_package_data = True,
        zip_safe = False,
        keywords = KEYWORDS,
        classifiers = CLASSIFIERS,
    )

    # run_build = RUN_CYTHON_BUILD
    # if run_build:
    #     generate_cython()

    setup(**metadata)


if __name__ == '__main__':
    # Running the build is as simple as: 
    # >> python setup.py sdist bdist_wheel
    # This command includes building the required Python extensions (Cython included)

    # It's recommended, however, to use:
    # >> python setup.py build_ext 
    # first, and then
    # >> python setup.py sdist bdist_wheel
    setup_package()


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

from ..io import imread 
from ..path import abspath, join
from .._base import __curr__ 
from ..annotations import Optional
import numpy as np 

from ..coreten import Tensor

here = join(__curr__, 'data').replace('\\', "/") + "/"

def _get_path_to_data(name) -> str:
    return join(here, name)


def audio_mixer(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of an audio mixer.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.audio_mixer()
        >> tens.shape
        (427, 640, 3)

    """
    return imread(here+'audio_mixer.jpg', target_size=target_size, rgb=rgb)


def bear(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a bear.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.bear()
        >> tens.shape
        (427, 640, 3)

    """
    return imread(here+'bear.jpg', target_size=target_size, rgb=rgb)


def beverages(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of beverages.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.beverages()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'beverages.jpg', target_size=target_size, rgb=rgb)


def black_cat(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a black cat.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.black_cat()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'black_cat.jpg', target_size=target_size, rgb=rgb)


def blue_tang(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x414 image Tensor (RGB, by default) of a blue tang (a type of fish).

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.audio_mixer()
        >> tens.shape
        (414, 640, 3)
        
    """
    return imread(here+'blue_tang.jpg', target_size=target_size, rgb=rgb)


def camera(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a camera.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.camera()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'camera.jpg', target_size=target_size, rgb=rgb)


def controller(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a game controller.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.controller()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'controller.jpg', target_size=target_size, rgb=rgb)


def drone(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x358 image Tensor (RGB, by default) of a robotic drone.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.drone()
        >> tens.shape
        (358, 640, 3)
        
    """
    return imread(here+'drone.jpg', target_size=target_size, rgb=rgb)


def dusk(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a dusk landscape.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.dusk()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'dusk.jpg', target_size=target_size, rgb=rgb)


def fighter_fish(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x640 image Tensor (RGB, by default) of a fighter fish.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.fighter_fish()
        >> tens.shape
        (640, 640, 3)
        
    """
    return imread(here+'fighter_fish.jpg', target_size=target_size, rgb=rgb)


def gold_fish(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x901 image Tensor (RGB, by default) of a gold fish.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.gold_fish()
        >> tens.shape
        (901, 640, 3)
        
    """
    return imread(here+'gold_fish.jpg', target_size=target_size, rgb=rgb)


def green_controller(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x512 image Tensor (RGB, by default) of a green game controller.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.green_controller()
        >> tens.shape
        (512, 640, 3)
        
    """
    return imread(here+'green_controller.jpg', target_size=target_size, rgb=rgb)


def green_fish(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x430 image Tensor (RGB, by default) of a green fish.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.green_fish()
        >> tens.shape
        (430, 640, 3)
        
    """
    return imread(here+'green_fish.jpg', target_size=target_size, rgb=rgb)


def guitar(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a guitar.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.guitar()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'guitar.jpg', target_size=target_size, rgb=rgb)


def island(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x426 image Tensor (RGB, by default) of an island.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.island()
        >> tens.shape
        (426, 640, 3)
        
    """
    return imread(here+'island.jpg', target_size=target_size, rgb=rgb)


def jellyfish(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a jellyfish.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.jellyfish()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'jellyfish.jpg', target_size=target_size, rgb=rgb)


def laptop(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a laptop.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.laptop()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'laptop.jpg', target_size=target_size, rgb=rgb)


def mountain(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a mountain.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.mountain()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'mountain.jpg', target_size=target_size, rgb=rgb)


def night(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a night landscape.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.night()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'night.jpg', target_size=target_size, rgb=rgb)


def puppies(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a litter of puppies.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.puppies()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'puppies.jpg', target_size=target_size, rgb=rgb)


def puppy(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x512 image Tensor (RGB, by default) of a puppy.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.puppy()
        >> tens.shape
        (512, 640, 3)
        
    """
    return imread(here+'puppy.jpg', target_size=target_size, rgb=rgb)


def red_fish(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a red fish.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.red_fish()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'red_fish.jpg', target_size=target_size, rgb=rgb)


def phone(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a rotary phone.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.phone()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'rotary_phone.jpg', target_size=target_size, rgb=rgb)


def sea_turtle(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x400 image Tensor (RGB, by default) of a sea turtle.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sea_turtle()
        >> tens.shape
        (400, 640, 3)
        
    """
    return imread(here+'sea_turtle.jpg', target_size=target_size, rgb=rgb)


def snow(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x360 image Tensor (RGB, by default) of snow.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.snow()
        >> tens.shape
        (360, 640, 3)
        
    """
    return imread(here+'snow.jpg', target_size=target_size, rgb=rgb)


def snowflake(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x480 image Tensor (RGB, by default) of a snowflake.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.snowflake()
        >> tens.shape
        (480, 640, 3)
        
    """
    return imread(here+'snowflake.jpg', target_size=target_size, rgb=rgb)


def sunrise(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a sunrise landscape.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.sunrise()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'sunrise.jpg', target_size=target_size, rgb=rgb)


def tent(target_size: Optional[tuple] = None, rgb: bool = True) -> Tensor:
    r"""
        Returns a standard 640x427 image Tensor (RGB, by default) of a tent.

    Args:
        target_size (Optional[tuple]): Intended target size (follows the ``(width, height)`` format).
            If None, the unaltered tensor will be returned.
        rgb (bool): Boolean whether to return an RGB Tensor (default is ``True``).
    
    Returns:
        Tensor of shape ``(height, width, channels)``.

    Examples::

        >> tens = caer.data.tent()
        >> tens.shape
        (427, 640, 3)
        
    """
    return imread(here+'tent.jpg', target_size=target_size, rgb=rgb)


__all__ = [d for d in dir() if not d.startswith('_')]
# __all__ = [
#     'audio_mixer',
#     'bear',
#     'beverages',
#     'black_cat',
#     'blue_tang',
#     'camera',
#     'controller',
#     'drone',
#     'dusk',
#     'fighter_fish',
#     'gold_fish',
#     'green_controller',
#     'green_fish',
#     'guitar',
#     'island',
#     'jellyfish',
#     'laptop',
#     'mountain',
#     'night',
#     'puppies',
#     'puppy',
#     'red_fish',
#     'phone',
#     'sea_turtle',
#     'snow',
#     'sunrise',
#     'tent'
# ]


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

import cv2 as cv 
import numpy as np 

from ..coreten import Tensor
from .position import _proc_in_chunks


__all__ = [
    'blur',
    'gaussian_blur',
    'median_blur',
    'motion_blur'
]

def blur(tens, ksize) -> Tensor:
    func = _proc_in_chunks(cv.blur, ksize=(ksize, ksize))
    return func(tens)


def gaussian_blur(tens, ksize, sigma=0) -> Tensor:
    # When sigma=0, it is computed as `sigma = 0.3*((ksize-1)*0.5 - 1) + 0.8`
    func = _proc_in_chunks(cv.GaussianBlur, ksize=(ksize, ksize), sigmaX=sigma)
    return func(tens)


def median_blur(tens, ksize) -> Tensor:
    if tens.dtype == np.float32 and ksize not in {3, 5}:
        raise ValueError(
            "Invalid ksize value {}. For a float32 image the only valid ksize values are 3 and 5".format(ksize)
        )

    func = _proc_in_chunks(cv.medianBlur, ksize=ksize)
    return func(tens)


def motion_blur(tens, kernel) -> Tensor:
    func = _proc_in_chunks(cv.filter2D, ddepth=-1, kernel=kernel)
    return func(tens)


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import cv2 as cv
import numpy as np 
from urllib.request import urlopen
# from urllib.error import URLError

from .resize import resize
from ..coreten import to_tensor, Tensor
from ..color import rgb2bgr, to_bgr
from ..path import exists
from .._internal import _check_target_size

__all__ = [
    'imread',
    'imsave'
]

IMREAD_COLOR = 1


def imread(image_path, rgb=True, target_size=None, resize_factor=None, preserve_aspect_ratio=False, interpolation='bilinear') -> Tensor:
    r"""
        Loads in an image from `image_path` (can be either a system filepath or a URL)

        Args:
            image_path (str): Filepath/URL to read the image from.
            rgb (bool): Boolean to keep RGB ordering. Default: True

        Returns:
            Tensor with shape ``(height, width, channels)``.

        Examples::

            >> tens = caer.imread(tens_path) # From FilePath
            >> tens.shape
            (427, 640, 3)

            >> tens = caer.imread('https://raw.githubusercontent.com/jasmcaus/caer/master/caer/data/beverages.jpg') # From URL
            >> tens.shape
            (427, 640, 3)

    """    

    return _imread(image_path, rgb=rgb,target_size=target_size, resize_factor=resize_factor, preserve_aspect_ratio=preserve_aspect_ratio, interpolation=interpolation)


def _imread(image_path, rgb=True, target_size=None, resize_factor=None, preserve_aspect_ratio=False, interpolation='bilinear') -> Tensor:   
    if target_size is not None:
        _ = _check_target_size(target_size)
    
    # if not isinstance(channels, int) or channels not in [1, 3]:
    #     raise ValueError('channels must be an integer - 1 (Grayscale) or 3 (RGB)')

    interpolation_methods = {
        'nearest':  0, '0': 0,  0: 0, # 0
        'bilinear': 1, '1': 1,  1: 1, # 1
        'bicubic':  2, '2': 2,  2: 2, # 2
        'area':     3, '3': 3,  3: 3  # 3
    }

    if interpolation not in interpolation_methods:
        raise ValueError('Specify a valid interpolation type - area/nearest/bicubic/bilinear')


    if exists(image_path):
        tens = _read_image(image_path) # returns RGB

    # TODO: Create URL validator
    elif image_path.startswith(('http://', 'https://')):
        # Returns RGB image
        tens = _url_to_image(image_path)
        
        # If the URL is valid, but no image at that URL, NoneType is returned
        if tens is None:
            raise ValueError('The URL specified does not point to an image')

    else:
        raise ValueError('Specify either a valid URL or filepath')

    # try:
    #     # Returns RGB image
    #     tens = _url_to_image(image_path)
        
    #     # If the URL is valid, but no image at that URL, NoneType is returned
    #     if tens is None:
    #         raise ValueError('The URL specified does not point to an image')

    #     # return tens

    # # If the URL is invalid
    # except (Exception, URLError):
    #     if exists(image_path):
    #         tens = _read_image(image_path) # returns RGB

    #     else:
    #         raise ValueError('Specify either a valid URL or filepath')
    
    if target_size is not None or resize_factor is not None:
        # Enforce a Tensor is passed to resize() 
        to_tensor(tens, cspace='rgb')
        tens = resize(tens, target_size, resize_factor=resize_factor, preserve_aspect_ratio=preserve_aspect_ratio,interpolation=interpolation)


    # If `rgb=False`, then we assume that BGR is expected
    if not rgb:
        tens = rgb2bgr(tens)
        # We need to convert back to tensor
        return to_tensor(tens, cspace='bgr')
    
    return to_tensor(tens, cspace='rgb')


def _read_image(image_path):
    r"""
        Returns an RGB ndarray
    """
    if not exists(image_path):
        raise FileNotFoundError('The image file was not found')

    # BGR image
    tens =  cv.imread(image_path)

    # Convert to RGB
    # WARNING: DO NOT USE to_rgb() as it creates a brand new Tensor (which defaults to RGB)
    # This issue will, hopefully, be fixed in a future update.
    # tens = to_rgb(tens)
    return cv.cvtColor(tens, cv.COLOR_BGR2RGB)


def _url_to_image(url):
    r"""
        Returns an RGB ndarray.
    """
    response = urlopen(url)
    tens = np.asarray(bytearray(response.read()), dtype='uint8')
    # BGR image
    tens = cv.imdecode(tens, IMREAD_COLOR)

    if tens is not None:
        return cv.cvtColor(tens, cv.COLOR_BGR2RGB)
        
    else:
        raise ValueError(f'No image found at "{url}"')


def imsave(path, tens) -> bool:
    r"""
        Saves a Tensor to `path`
            
    Args:
        path (str): Filepath to save the image to 
    
    Returns
        ``True``; if `tens` was written to `path`
        ``False``; otherwise

    Examples::

        >> tens = caer.data.audio_mixer()
        >> caer.imsave('audio_mixer.png', tens)
        True

    """
    if not isinstance(tens, Tensor):
        raise TypeError('`tens` must be a caer.Tensor')

    # Convert to tensor 
    _ = tens._nullprt() # raises a ValueError if we're dealing with a Foreign Tensor with illegal `.cspace` value

    try:
        # OpenCV uses BGR Tensors and saves them as RGB images
        if tens.cspace !='bgr':
            tens = to_bgr(tens)
        return cv.imwrite(path, tens)
    except:
        raise ValueError('`tens` needs to be a caer Tensor. Try reading the image using `caer.imread()`. More support for additional platforms will follow. Check the Changelog for further details.')


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

r"""
    caer.transforms consists of position and color-based transforms for use.
"""

# Don't import from functional.py

from .position import (
    hflip,
    vflip,
    hvflip,
    rand_flip,
    scale,
    translate,
    transpose,
    rotate,
    crop,
    center_crop,
    rand_crop,
    posterize,
    solarize,
    equalize,
    clip,
    pad,
    __all__ as __all_pos__
)

from .color import (
    adjust_brightness,
    adjust_contrast,
    adjust_hue,
    adjust_saturation,
    adjust_gamma,
    affine,
    darken,
    brighten,
    random_brightness,
    correct_exposure,
    augment_random,
    __all__ as __all_color__
)

from .simulate import (
    sim_snow,
    sim_rain,
    sim_fog,
    sim_gravel,
    sim_sun_flare,
    sim_motion_blur,
    sim_autumn,
    sim_shadow,
    __all__ as __all_sim__
)

__all__ = __all_color__ + __all_pos__ + __all_sim__


# Stop polluting the namespace
del __all_pos__
del __all_color__
del __all_sim__


#    _____           ______  _____
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


class _TensorBase:
    def __init__(self):
        self.max_width = 2
        self.is_floating_point = False
        self.sci_mode = False
        self.int_mode = True
        self.numelem = self.size
        self.foreign = True

        # self._mode = 'rgb'
        # self.mode = self._mode # '._mode' is used internally --> prevents misuse of the API
        self.cspace = 'null' # default


    def __repr__(self):
        return "<class 'caer.Tensor'>"

    def __str__(self):
        return self.__repr__()

    def height(self):
        return self.shape[0]

    def width(self):
        return self.shape[1]

    def channels(self):
        return self.shape[2]

    def cmode(self):
        return self.cspace

    def numel(self):
        return self.size

    def dim(self):
        return self.ndim

    def size_dim(self, dim):
        return self.shape[dim]

    def type(self):
        return self.dtype

    def clone(self):
        return self.copy()

    # Colorspace stuff
    def is_rgb(self):
        return self.cspace == 'rgb'

    def is_bgr(self):
        return self.cspace == 'bgr'

    def is_gray(self):
        return self.cspace == 'gray'

    def is_hsv(self):
        return self.cspace == 'hsv'

    def is_lab(self):
        return self.cspace == 'lab'

    def is_hls(self):
        return self.cspace == 'hls'

    def is_yuv(self):
        return self.cspace == 'yuv'

    def is_luv(self):
        return self.cspace == 'luv'

    # Foreign Tensor-stuff
    def is_foreign(self):
        return self.foreign

    def _is_valid_cspace(self):
        if (self.cspace == 'rgb') or (self.cspace == 'bgr') or (self.cspace == 'gray') or (self.cspace == 'hsv') or (self.cspace == 'hls') or (self.cspace == 'lab') or (self.cspace == 'yuv') or (self.cspace == 'luv'):
            return True

        # Else
        self.cspace = 'null'
        return False


    def is_null(self):
        r"""
            Returns True if the ``.cspace`` attribute is valid (either bgr/rgb/gray/hsv/hls/lab)
            Returns False, otherwise (usually happens when foreign arrays (like ndarrays) are converted to Caer Tensors).
        """

        return not self._is_valid_cspace()


    def _nullprt(self):
        r"""
            NOT nullptr in C/C++.
            Raises a TypeError ==> usually happens when foreign arrays (like ndarrays) are converted to Caer Tensors.
        """
        if self.is_null():
            raise TypeError('IllegalTensorWarning: Cannot determine the colorspace for this foreign tensor. You can set it manually by modifying the `.cspace` attribute. We suggest operating solely in Caer Tensors.')



#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>

# Resize
from .resize import (
    resize,
    smart_resize,
    __all__ as __all_res__
)

# Image 
from .io import (
    imread,
    imsave,
    __all__ as __all_io__
)

__all__ = __all_io__ + __all_res__


#    _____           ______  _____ 
#  / ____/    /\    |  ____ |  __ \
# | |        /  \   | |__   | |__) | Caer - Modern Computer Vision
# | |       / /\ \  |  __|  |  _  /  Languages: Python, C, C++, Cuda
# | |___   / ____ \ | |____ | | \ \  http://github.com/jasmcaus/caer
#  \_____\/_/    \_ \______ |_|  \_\

# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2020-2021 The Caer Authors <http://github.com/jasmcaus>


import math
import time
import cv2 as cv

from .._internal import _check_target_size
from ..path import list_videos, exists, mkdir
from .constants import FRAME_COUNT, FPS
from ..io import imsave, resize

__all__ = [
    'extract_frames'
]

def extract_frames(input_folder, 
                   output_folder, 
                   target_size=None, 
                   recursive=False,
                   label_counter = None, 
                   max_video_count=None, 
                   frames_per_sec=None, 
                   frame_interval=None,
                   dest_filetype='jpg') -> int:
    r"""
        Extract frames from videos within a directory and save them as separate frames in an output directory.

    Args:
        input_folder (str): Input video directory.
        output_folder (str): Output directory to save the frames.
        target_size (tuple): Destination Image Size (tuple of size 2)
        label_counter (int): Starting label counter (optional)
        max_video_count (int): Number of videos to process.
        frames_per_sec (int, float): Number of frames to process per second. 
        frame_interval (int, float): Interval between the frames to be processed.
        dest_filetype (str): Processed image filetype (png, jpg). Default: png

    Returns:
        label_counter (after processing)

    """

    dest_filetype.replace('.', '')
    processed_videos = 0
    vid_count = 0 # to check if < max_video_count

    if not exists(input_folder):
        raise ValueError('Input folder does not exist', input_folder)
    
    if target_size is not None:
        _ = _check_target_size(target_size)
    
    video_list = list_videos(input_folder, recursive=recursive, use_fullpath=True, verbose=0)

    if len(video_list) == 0:
        raise ValueError(f'No videos found at {input_folder}')

    if label_counter is None:
        label_counter = 0

    if max_video_count is None:
        max_video_count = len(video_list)

    if not exists(output_folder):
        mkdir(output_folder)

    # Begin Timer
    start = time.time()

    for vid_filepath in video_list:
        if vid_count < max_video_count:
            capture = cv.VideoCapture(vid_filepath)
            video_frame_counter = 0
            vid_count += 1

            # Find the number of frames and FPS
            video_frame_count = int(capture.get(FRAME_COUNT)) - 1
            video_fps = math.ceil(capture.get(FPS))
            file = vid_filepath[vid_filepath.rindex('/')+1:]
            
            if frames_per_sec is not None:
                if frame_interval is None:
                    interval = _determine_interval(video_fps/frames_per_sec) # eg: 30//15
            
                else:
                    interval = frame_interval

            # if frames_per_sec and frame_interval are both None, we assume that each frame should be processed
            else:
                interval = 1
            
            # processed_frames = (video_frame_count//video_fps) * frames_per_sec

            print(f'{vid_count}. Reading \'{file}\'. Frame Count: {video_frame_count}. FPS: {video_fps}. Processed frames: {video_frame_count//interval}')
            
            # Start converting the video
            while capture.isOpened():
                _, frame = capture.read()

                if target_size is not None:                    
                    frame = resize(frame, target_size=target_size)
                
                # Write the results back to output location as per specified frames per second
                if video_frame_counter % interval == 0:
                    imsave(f'{output_folder}/{file}_{label_counter}.{dest_filetype}', frame)
                    video_frame_counter += 1
                    label_counter += 1
                    # print('Frame counter: ', video_frame_counter)
                
                video_frame_counter += 1

                # If there are no more frames left
                if video_frame_counter > (video_frame_count-1):
                    capture.release()
                    processed_videos += 1
                    break
    # End timer
    end = time.time()
    
    # Printing stats
    taken = end-start
    minu = taken // 60
    sec = taken % 60
    if processed_videos > 1:
        print(f'[INFO] {processed_videos} videos extracted in {minu:.0f}m {sec:.0f}s')
    else:
        print(f'[INFO] {processed_videos} video extracted in {minu:.0f}m {sec:.0f}s')

    return label_counter


def _determine_interval(x) -> int:
    y = '{x:.1f}'
    inde = y.find('.') + 1
    if inde == -1: # if no '.' (if an integer)
        return x
    if int(y[inde]) < 5:
        return math.floor(x)
    else:
        return math.ceil(x)
